/*****************************************************************************/
//
// FILE        :  Image_fft.cc
// AUTHOR      :  Jacco Vink 
// DESCRIPTION :  Implementation of Image class (contains routines which use
//                 fftw)
//
//
// COPYRIGHT   :  Jacco Vink, AIP, 2000
//
/*****************************************************************************/
#include "Image.h"


/*****************************************************************************/
//
// FUNCTION    : fft()
// DESCRIPTION : Uses fftw routines to get the Fourier transform of the image.
//               Afterwards the fftw_plan is set to the reversed direction.
// 
//
// REMARKS     :
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::fft( void )
{

  // only changes anything when complex array does not exist!:
  if( allocate_complex() == FAILURE ){ 
    // warning("fft(): Could not allocate memory for complex array!");
    return FAILURE;
  }

  if( fftwdir ==  FFTW_COMPLEX_TO_REAL || fftplanexist == false){
    float scale= 1.0/(nx*ny);
    if( fftplanexist == false)
      fftplan =rfftw2d_create_plan(ny, nx,FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);

    rfftwnd_one_complex_to_real(fftplan, &cpixel[0], &pixel[0] );
    for( unsigned long i=0; i < (unsigned long)nx*ny ; i++)
      pixel[i] *= scale;

    ftransformed = true;
    rfftwnd_destroy_plan( fftplan);
    fftplan = rfftw2d_create_plan(ny, nx, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
    fftwdir = FFTW_REAL_TO_COMPLEX;
    fftplanexist = true;
  }
  else{
    rfftwnd_one_real_to_complex(fftplan, &pixel[0], &cpixel[0] );
    ftransformed = true;
    rfftwnd_destroy_plan( fftplan);
    fftplan = rfftw2d_create_plan(ny, nx, FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE);
    fftwdir =  FFTW_COMPLEX_TO_REAL;
    fftplanexist = true;
  }//if

  return SUCCESS;
}// fft()

/*****************************************************************************/
//
// FUNCTION    : convolve()
// DESCRIPTION : Convolves the image with another image using FFTW.
//               The fft plan for "psf" has to be the backward transform.
//               It will call fft() and perform a forward transform, when
//               this is not the case.
//               Expect :  psf->fftwdir =  FFTW_COMPLEX_TO_REAL
//                        this->fftwdir =  FFTW_REAL_TO_COMPLEX
//
// 
//
// REMARKS     :
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::convolve( Image *psf, bool newfft )
{

  if( nx != psf->nx || ny != psf->ny || nx <= 0 || ny <= 0){
    // warning("convolve( ): Input images should be of the same size!");
    return FAILURE;
  }//

  if( allocate_complex(FFTW_COMPLEX_TO_REAL ) == FAILURE ){
    // warning("convolve(): Could not allocate memory for complex array!");
    return FAILURE;
  }//if

  if( psf->fftwdir != FFTW_REAL_TO_COMPLEX ){
    puts("Change psf!");
     rfftwnd_destroy_plan( psf->fftplan);
     psf->fftplan = rfftw2d_create_plan(ny, nx,FFTW_REAL_TO_COMPLEX, 
					FFTW_ESTIMATE);
  }//if

  if( newfft == true || psf->ftransformed == false ){
    if( psf->allocate_complex() == FAILURE )
      return FAILURE;
    rfftwnd_one_real_to_complex(psf->fftplan, &(psf->pixel[0]), 
				&(psf->cpixel[0] ));
    psf->ftransformed = true; 
  }//if

  rfftwnd_one_real_to_complex(psf->fftplan, &pixel[0], &cpixel[0] );
  ftransformed = true;

  float scale= 1.0/(nx*ny), newre, newim;
  int pixnum;
  for(unsigned int j=0 ; j < ny ; j++)
    for(unsigned int i=0 ; i <= nx/2 ; i++){
      pixnum = j*(nx/2+1) + i;

      newre = cpixel[pixnum].re * psf->cpixel[pixnum].re - 
       	      cpixel[pixnum].im * psf->cpixel[pixnum].im;
      newim = cpixel[pixnum].re * psf->cpixel[pixnum].im +
	      cpixel[pixnum].im * psf->cpixel[pixnum].re;
      cpixel[pixnum].re = newre*scale;
      cpixel[pixnum].im = newim*scale;

    }//for

  rfftwnd_one_complex_to_real( fftplan, &cpixel[0], &pixel[0]);
  ftransformed = false;

  return SUCCESS;
}//convolve

/*****************************************************************************/
//
// FUNCTION    : convolve()
// DESCRIPTION : Convolves the image with another image using FFTW
//               The fftw_plans should be as followed:
//               im -> contains backward transform plan
//               psf -> should contain forward transform plan
//               For the psf this is checked and corrected, but the speed
//               should be faster when things are right from the beginning.
//
// REMARKS     :
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::convolve(const Image &im, Image *psf, bool newfft )
{

  if( im.nx != psf->nx || im.ny != psf->ny || im.nx <= 0 || im.ny <= 0){
    // warning("convolve( ): Input images should be of the same size!");
    return FAILURE;
  }//



  if( im.ftransformed == false || im.ncomplex < 2 || 
      im.fftwdir != FFTW_COMPLEX_TO_REAL ){
    // warning("convolve(): Error in FFTW direction or complex part of the input image!");
    puts("convolve(): Error in FFTW direction or complex part of the input image!");
    return FAILURE;
  }//if


  if( nx != im.nx || ny != im.ny)
    if( allocate( im.nx, im.ny)== FAILURE )
      return FAILURE;


  if( allocate_complex() == FAILURE ){
    // warning("convolve(): Could not allocate memory for complex array!");
    return FAILURE;
  }//if


  if( psf->fftwdir != FFTW_REAL_TO_COMPLEX ){
     rfftwnd_destroy_plan( psf->fftplan);
     psf->fftplan = rfftw2d_create_plan(ny, nx,FFTW_COMPLEX_TO_REAL, 
					FFTW_ESTIMATE);
  }//if


  if( newfft == true || psf->ftransformed == false ){
    if( psf->allocate_complex() == FAILURE )
      return FAILURE;
    rfftwnd_one_real_to_complex(psf->fftplan, &(psf->pixel[0]), 
				&(psf->cpixel[0] ));
    psf->ftransformed = true; 
  }//if


  float scale= 1.0/(nx*ny), newre, newim;
  int pixnum;
  for(unsigned int j=0 ; j < ny ; j++)
    for(unsigned int i=0 ; i < nx/2+1 ; i++){
      pixnum = j*(nx/2+1) + i;

      newre = im.cpixel[pixnum].re * psf->cpixel[pixnum].re - 
       	      im.cpixel[pixnum].im * psf->cpixel[pixnum].im;
      newim = im.cpixel[pixnum].re * psf->cpixel[pixnum].im +
	      im.cpixel[pixnum].im * psf->cpixel[pixnum].re;
      cpixel[pixnum].re = newre*scale;
      cpixel[pixnum].im = newim*scale;
    }//for

  rfftwnd_one_complex_to_real(im.fftplan, &cpixel[0], &pixel[0]);
  ftransformed = false;


  return SUCCESS;
}//convolve


/*****************************************************************************/
//
// FUNCTION    : makeGaussFT()
// DESCRIPTION : Makes the Fourier transform of a Gaussian.
//
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::makeGaussFT( const Image &im, float sigmax, float sigmay)
{
  if( sigmax <= 0.0 || sigmay <= 0.0 ){
    //    warning("makeGaussFT(): Incorrect input parameter");
    return FAILURE;
  }

  if( im.nx <= 0 || im.ny <= 0 ){
    //    warning("makeGaussFT(): Incorrect input image!");
    return FAILURE;
  }

  if( allocate(im.nx, im.ny) == FAILURE ){
    //    warning("makeGaussFT(): Allocation error!");
    return FAILURE;
  }//if

  if( allocate_complex(  FFTW_REAL_TO_COMPLEX ) == FAILURE ){
    //    warning("makeGaussFT(): Allocation error!");
    return FAILURE;
  }//if


  unsigned long offset1, offset2, nxc = (nx/2 + 1);
  double pisqr = M_PI * M_PI;
  double kx, ky, kxsigma2, kysigma2;
  float value;

  if( nx != ny || sigmax != sigmay ){
    for( unsigned int j = 0 ; j < ny/2 +1 ; j++){
      ky = (double)j/ny;          
      kysigma2 = SQR(ky*sigmay);
      offset1 = nxc*j;
      offset2 = nxc*(ny-j);

      for( unsigned int i=0; i < nx/2 +1; i++ ){
	
	kx = (double)i/nx;
	kxsigma2 = SQR(kx*sigmax);
	value = float( exp( -2*pisqr*(kxsigma2 + kysigma2)  ));
	cpixel[i + offset1].re = value;
	cpixel[i + offset1].im = 0.0;
	if( j > 0 && j != ny/2 ){
	  cpixel[i + offset2].re = value;
	  cpixel[i + offset2].im = 0.0;
	}

      }//for
    }//for
  }
  else{ // make use of the extra symmetry
    for( unsigned int j = 0 ; j < ny/2 +1 ; j++){
      ky = (double)j/ny;          
      kysigma2 = SQR(ky*sigmay);
      offset1 = nxc*j;
      offset2 = nxc*(ny-j);

      for( unsigned int i=j; i < nx/2 + 1; i++ ){
	kx = (double)i/nx;
	kxsigma2 = SQR(kx*sigmax);
	value = float( exp( -2*pisqr*(kxsigma2 + kysigma2)  ));
	cpixel[i + offset1].re = value;
	cpixel[i + offset1].im = 0.0;
	if( j > 0 && j != ny/2 ){
	  cpixel[i + offset2].re = value;
	  cpixel[i + offset2].im = 0.0;
	}//if
	if( i > j  && j < nxc && i <= ny/2 ){
	  // use symmetry and interchange i and j (x and y)
	  cpixel[j + i*nxc].re = value;
	  cpixel[j + i*nxc].im = 0.0;
	  if( i > 0 && i != ny/2 ){
	    cpixel[j + nxc*(ny-i)].re = value;
	    cpixel[j + nxc*(ny-i)].im = 0.0;
	  }//if
	}//if
	
      }//for
    }//for
  }//else

  ftransformed = true;

  return SUCCESS;

}//makeGaussFT()


/*****************************************************************************/
//
// FUNCTION    : mexicanhatFT()
// DESCRIPTION : Makes the Fourier transform of the mexican hat.
//               Note that only one quadrant needs to be computed.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    :
//
/*****************************************************************************/
Status Image::mexicanhatFT(const Image &im, float sigmax, float sigmay )
{
  if( im.nx <= 0 || im.ny <= 0 || sigmax <= 0.0 || sigmay <= 0.0 ){
    //    warning("mexicanhatFT(): Invalid input parameters.");
    return FAILURE;
  }// 

  if( allocate(im.nx, im.ny) == FAILURE ){
    //    warning("mexicanhatFT(): Allocation error!");
    return FAILURE;
  }//if

  if( allocate_complex(  FFTW_REAL_TO_COMPLEX ) == FAILURE ){
    //  warning("mexicanhatFT(): Allocation error!");
    return FAILURE;
  }//if


  unsigned long offset1, offset2, nxc = (nx/2 + 1);
  double fac1 = 2.0, fac2, fac3= 2.0 * M_PI * M_PI;
  double kx, ky, kxsigma2, kysigma2;
  float value;

  if( nx != ny || sigmax != sigmay ){
    for( unsigned int j = 0 ; j < ny/2 +1 ; j++){
      ky = (double)j/ny;          
      kysigma2 = SQR(ky*sigmay);
      offset1 = nxc*j;
      offset2 = nxc*(ny-j);

      for( unsigned int i=0; i < nx/2 +1; i++ ){
	
	kx = (double)i/nx;
	kxsigma2 = SQR(kx*sigmax);
	fac2 = fac3 * (kxsigma2 + kysigma2);
	value = float(fac1 * fac2 * exp( -fac2 ) );
	cpixel[i + offset1].re = value;
	cpixel[i + offset1].im = 0.0;
	if( j > 0 && j != ny/2 ){
	  cpixel[i + offset2].re = value;
	  cpixel[i + offset2].im = 0.0;
	}

      }//for
    }//for
  }
  else{ // make use of the extra symmetry
    for( unsigned int j = 0 ; j < ny/2 +1 ; j++){
      ky = (double)j/ny;          
      kysigma2 = SQR(ky*sigmay);
      offset1 = nxc*j;
      offset2 = nxc*(ny-j);

      for( unsigned int i=j; i < nx/2 + 1; i++ ){
	kx = (double)i/nx;
	kxsigma2 = SQR(kx*sigmax);
	fac2 = fac3 * (kxsigma2 + kysigma2);
	value = float(fac1 * fac2 * exp( -fac2 ) );
	cpixel[i + offset1].re = value;
	cpixel[i + offset1].im = 0.0;
	if( j > 0 && j != ny/2 ){
	  cpixel[i + offset2].re = value;
	  cpixel[i + offset2].im = 0.0;
	}//if
	if( i > j  && j < nxc && i <= ny/2 ){
	  // use symmetry and interchange i and j (x and y)
	  cpixel[j + i*nxc].re = value;
	  cpixel[j + i*nxc].im = 0.0;
	  if( i > 0 && i != ny/2 ){
	    cpixel[j + nxc*(ny-i)].re = value;
	    cpixel[j + nxc*(ny-i)].im = 0.0;
	  }//if
	}//if
	
      }//for
    }//for
  }//else

  ftransformed = true;


  return SUCCESS;

}// mexicanhatFT()

/*****************************************************************************/
//
// FUNCTION    : makeRingFT()
// DESCRIPTION : Makes the Fourier transform of the function 
//               4pi sigma^2r^2exp(-0.5r^2/sigma^2)
//
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status Image::makeRingFT( const Image &im, float sigmax, float sigmay)
{
  if( sigmax <= 0.0 || sigmay <= 0.0 ){
    //    warning("makeRingFT(): Incorrect input parameter");
    return FAILURE;
  }

  if( im.nx <= 2 || im.ny <= 2 ){
    //    warning("makeRingFT(): Incorrect input image!");
    return FAILURE;
  }

  if( allocate(im.nx, im.ny) == FAILURE ){
    //    warning("makeRingFT(): Allocation error!");
    return FAILURE;
  }//if

  if( allocate_complex(  FFTW_REAL_TO_COMPLEX ) == FAILURE ){
    //    warning("makeRingFT(): Allocation error!");
    return FAILURE;
  }//if


  unsigned long offset1, offset2, nxc = (nx/2 + 1);
  double pisqr = M_PI * M_PI;
  double kx, ky, kxsigma2, kysigma2;
  float value;

  if( nx != ny || sigmax != sigmay ){
    for( unsigned int j = 0 ; j < ny/2 +1 ; j++){
      ky = (double)j/ny;          
      kysigma2 = SQR(ky*sigmay);
      offset1 = nxc*j;
      offset2 = nxc*(ny-j);

      for( unsigned int i=0; i < nx/2 +1; i++ ){
	
	kx = (double)i/nx;
	kxsigma2 = SQR(kx*sigmax);
	value = float( (1.0 - 2.0*pisqr*(kxsigma2 + kysigma2))*
		       exp( -2*pisqr*(kxsigma2 + kysigma2)  ));
	cpixel[i + offset1].re = value;
	cpixel[i + offset1].im = 0.0;
	if( j > 0 && j != ny/2 ){
	  cpixel[i + offset2].re = value;
	  cpixel[i + offset2].im = 0.0;
	}

      }//for
    }//for
  }
  else{ // make use of the extra symmetry
    for( unsigned int j = 0 ; j < ny/2 +1 ; j++){
      ky = (double)j/ny;          
      kysigma2 = SQR(ky*sigmay);
      offset1 = nxc*j;
      offset2 = nxc*(ny-j);

      for( unsigned int i=j; i < nx/2 + 1; i++ ){
	kx = (double)i/nx;
	kxsigma2 = SQR(kx*sigmax);
	value = float( (1.0 - 2.0*pisqr*(kxsigma2 + kysigma2))*
		       exp( -2*pisqr*(kxsigma2 + kysigma2)  ));
	cpixel[i + offset1].re = value;
	cpixel[i + offset1].im = 0.0;
	if( j > 0 && j != ny/2 ){
	  cpixel[i + offset2].re = value;
	  cpixel[i + offset2].im = 0.0;
	}//if
	if( i > j  && j < nxc && i <= ny/2 ){
	  // use symmetry and interchange i and j (x and y)
	  cpixel[j + i*nxc].re = value;
	  cpixel[j + i*nxc].im = 0.0;
	  if( i > 0 && i != ny/2 ){
	    cpixel[j + nxc*(ny-i)].re = value;
	    cpixel[j + nxc*(ny-i)].im = 0.0;
	  }//if
	}//if
	
      }//for
    }//for
  }//else

  ftransformed = true;

  return SUCCESS;

}//makeRingFT()
