#include "Dal.h"
#include "Error.h"
#include "errstr.h"
#include "Image.h"
#include "srclist.h"


Image bgdfilter, temp;


/*****************************************************************************/
//
// FUNCTION    : calcBackground()
// DESCRIPTION : Calculates the background image. An image containing the
//               detected sources is subtracted from the original image and
//               then the resulting image is convolved with a function of
//               the form r^exp(-0.5r^2/sigma^2). Sigma being twice the wavelet
//               scale.
// 
//
// SIDE EFFECTS:
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//     background        O                background image
//     inputImage        I                original image
//     reconstruction    I                image containing the detected 
//                                          sources (may be zero).
//     expmask           I                image containging only 0 and 1,
//                                          1 for those places where a source
//                                          can be detected, 0 for parts 
//                                          outside the detector
//     scale             I                 wavelet scale for which to calculate
//                                           background image.
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status calcBackground( Image *background,
		       const Image &inputImage, 
		       const Image &reconstruction, 
		       const Image &expmask, 
		       float scale)
{


  *background = inputImage;
  *background -= reconstruction;



  if( bgdfilter.makeRingFT( *background, 2.0*scale, 2.0*scale )== FAILURE ||
      background->convolve( &bgdfilter, false ) == FAILURE){
    //    Sas::warning("bgdFilter","Error in filtering new background image!");
    return FAILURE;
  }


  //
  // correct for sharp edges by convolving the mask (with edges) similar
  // to the background and divide the background with this.
  // expmask should only contain 0 and 1

  temp=expmask;
  if( temp.convolve( &bgdfilter, false ) == FAILURE){
    //    Sas::warning("bgdFilter","Error in filtering new background image!");
    return FAILURE;
  }//if


  *background /= temp;

  // set negative pixels to 0:
  

  if( background->clip(0.0, 1.1*background->maximum() ) == FAILURE){
    //    Sas::warning("bgdClip","Error clipping new background image!");
    return FAILURE;
  }

  *background *= expmask;

 //  errstr << "set min background to: " << 0.2*background->mean()  << 
 //    Sas::message(Msg::App, Msg::Verbose);
 //
 // if( background->cage(0.2*background->mean(), 1.1*background->maximum() ) == FAILURE){
    //    Sas::warning("bgdClip","Error clipping new background image!");
 //    return FAILURE;
 //  }
 //
 //  *background *= expmask;


  return SUCCESS;
}//calcBackground()
