#include "Image.h"
#include "srclist.h"

/*****************************************************************************/
//
// FUNCTION    : reconstruct_image()
// DESCRIPTION : Makes an image which contains gaussian sources with the 
//               appropiate properties takene from the source list.
//               Uses the countrates for the normalization, bu the final image
//               is multiplied by the exposure map in order to give counts in
//               the final image.
// 
//
// SIDE EFFECTS: Assumes the list is ordered according to scale size.
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//     image              I                       input image
//     expmap             I                       exposure map
//     srclist            I                       source list
//     reconstruction     O                       reconstructed image
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status reconstructImage( const Image &image, const Image &expmap,
			 const Sourcelist &srclist, Image *reconstruction)
{

  float x, y, sigma, norm;
  *reconstruction = image;
  *reconstruction *= 0.0;

  if( srclist.ndetect == 0 ){
    return FAILURE;
  }// if


  for( unsigned int i = 0; i < srclist.ndetect ; i++ ){
    if( srclist.ignore[i] == false ){
      x = srclist.xima[i];
      y = srclist.yima[i];
      sigma = srclist.sigma[i];
      norm   = srclist.cntrate[i];
      reconstruction->addgaussian( x, y, norm, sigma, sigma, 0.0 );
    }//

  }//for

  *reconstruction *= expmap;

  return SUCCESS;

}// reconstructImage()



/*****************************************************************************/
//
// FUNCTION    : reconstructImage()
// DESCRIPTION : Makes an image which contains gaussian sources with the 
//               appropiate properties takene from the source list.
//               Only those sources are used which have scale < maxscale.
//               Uses the counts rather than the countrates for the 
//               normalizations.
//
// SIDE EFFECTS: Assumes the list is ordered according to scale size.
//               Sources which are put in the image are flagged "ignore=true"
//
// PARAMETERS  :
// 	NAME		I/O			DESCRIPTION
//     srclist            I                       source list
//     maxscale           I                       maximum scale size to use
//     reconstruction     I/O                     reconstructed image
//
//
//  RETURNS    : SUCCESS or FAILURE
//
/*****************************************************************************/
Status reconstructImage( const Sourcelist &srclist, float maxscale, 
			 Image *reconstruction)
{

  float x, y, sigma, norm;
   

  for( unsigned int i = 0; i < srclist.ndetect ; i++ ){

    if( srclist.ignore[i] == false && 0.99*srclist.scale[i] <= maxscale ){
      x = srclist.xima[i];
      y = srclist.yima[i];
      sigma = srclist.sigma[i];
      norm   = srclist.cnts[i];
      srclist.ignore[i] = true;
      reconstruction->addgaussian( x, y, norm, sigma, sigma, 0.0 );
    }//

  }//for


  return SUCCESS;

}// reconstructImage()
